/*
    getpkg

    getpkg <tool_name>
        - installs the specified tool (same as getpkg install <tool_name>)
        - confirms getpkg is fully initialised:
            - adds a line to the user's .bashrc file to source the getpkg ~/.bashrc_dropshell_tool script, if it doesn't exist
            - creates an empty ~/.bashrc_dropshell_tool script if it doesn't exist
            - creates the ~/.config/getpkg directory, if it does not exist
            - creates the ~/.local/bin/getpkg directory, if it does not exist

        - checks if ~/.config/getpkg/tool_name.json file exists. If it does, it will update the tool if the version is older than the remote version. If it doesn't exist, it will install the tool.
            - removes the tool from the user's system if it is already installed, and the tool's entries in the ~/.bashrc_dropshell_tool script
            - downloads the tool archive (tgz) from tools.dropshell.app (tool_name:ARCH), where ARCH is the architecture of the user's system, and unpacks it to the new tool directory: ~/.local/bin/getpkg/<tool_name>/
            - sets the PATH to include the tool directory, by modifying the ~/.bashrc_dropshell_tool script
            - adds an entry for autocompletion for the tool to the ~/.bashrc_dropshell_tool script, where autocomplete just runs <tool_name> autocomplete <args>
            - creates a ~/.config/getpkg/tool_name.json file, which contains the tool's name, version (by running <tool_name> version), hash from tools.dropshell.app, and architecture
            - if setup_script.sh exists, run the script

    getpkg install <tool_name>
        - legacy syntax for installing a tool (same as getpkg <tool_name>)

    getpkg publish <tool_name:ARCH> <folder>
        - creates a tgz archive of the folder, and uploads it to tools.dropshell.app, a simple object server.
        - prints the URL and hash of the uploaded archive
        - uses the token from env variable SOS_WRITE_TOKEN to write to tools.dropshell.app
        
    getpkg update
        - updates getpkg itself, and then runs install on all already installed tools

    getpkg autocomplete <args>
        - shows autocomplete for getpkg, and then exits
        - the tool list to choose from when calling install is hard coded in the autocomplete function

    getpkg version
        - prints the version of getpkg

    getpkg create <tool_name> <directory_name>
        - creates a new tool source directory in relative path <directory_name> if it doesn't exist
        - creates a setup_script.sh file in the tool source directory if it doesn't exist, that just prints a completion message and exits

    getpkg help
        - shows this help message


*/
#include "version.hpp"
#include "BashrcEditor.hpp"
#include "DropshellScriptManager.hpp"
#include "GetbinClient.hpp"
#include "ArchiveManager.hpp"
#include <iostream>
#include <string>
#include <vector>
#include <filesystem>
#include <fstream>
#include <cstring>
#include <nlohmann/json.hpp>

namespace {
using json = nlohmann::json;

// Compare versions (returns true if v1 < v2)
bool isVersionOlder(const std::string& v1, const std::string& v2) {
    // Simple version comparison - assumes versions are in YYYY.MMDD.HHMM format
    // or semantic versioning format (e.g., 1.2.3)
    return v1 < v2;
}

// Validate tool name to prevent path traversal and other issues
bool isValidToolName(const std::string& name) {
    if (name.empty() || name.length() > 100) return false;
    
    // Check for path traversal attempts
    if (name.find("..") != std::string::npos) return false;
    if (name.find("/") != std::string::npos) return false;
    if (name.find("\\") != std::string::npos) return false;
    
    // Only allow alphanumeric, dash, underscore, and dot
    for (char c : name) {
        if (!std::isalnum(c) && c != '-' && c != '_' && c != '.') {
            return false;
        }
    }
    
    return true;
}

// Escape string for shell command
std::string shellEscape(const std::string& str) {
    std::string result;
    result.reserve(str.length() + 20);
    result += "'";
    for (char c : str) {
        if (c == '\'') {
            result += "'\\''";
        } else {
            result += c;
        }
    }
    result += "'";
    return result;
}

std::string get_arch() {
#if defined(__x86_64__) || defined(_M_X64)
    return "x86_64";
#elif defined(__aarch64__)
    return "aarch64";
#else
    return "unknown";
#endif
}

std::string get_home() {
    const char* home = getenv("HOME");
    return home ? std::string(home) : std::string("");
}

int install_tool(int argc, char* argv[]) {
    if (argc < 3) {
        std::cerr << "Usage: getpkg install <tool_name>" << std::endl;
        return 1;
    }
    std::string toolName = argv[2];
    
    // Validate tool name
    if (!isValidToolName(toolName)) {
        std::cerr << "Invalid tool name. Tool names can only contain letters, numbers, dash, underscore, and dot." << std::endl;
        return 1;
    }
    
    std::string arch = get_arch();
    std::string home = get_home();
    std::filesystem::path configDir = std::filesystem::path(home) / ".config/getpkg";
    std::filesystem::path binDir = std::filesystem::path(home) / ".local/bin/getpkg" / toolName;
    std::filesystem::path archivePath = configDir / (toolName + ".tgz");
    std::filesystem::path toolInfoPath = configDir / (toolName + ".json");
    
    // Initialize directories
    std::filesystem::create_directories(configDir);
    std::filesystem::create_directories(binDir.parent_path());
    
    // Initialize bashrc
    dropshelltool::BashrcEditor bashrcEditor;
    bashrcEditor.addSourceLine();
    
    // Check if tool needs update or install
    bool needsUpdate = false;
    if (std::filesystem::exists(toolInfoPath)) {
        // Tool exists, check if update needed
        std::ifstream tfile(toolInfoPath);
        json toolInfo;
        tfile >> toolInfo;
        tfile.close();
        
        std::string localHash = toolInfo.value("hash", "");
        
        // Get remote hash to compare
        GetbinClient getbin;
        std::string remoteHash;
        if (getbin.getHash(toolName, arch, remoteHash) && !remoteHash.empty()) {
            if (localHash != remoteHash) {
                needsUpdate = true;
                std::cout << "Updating " << toolName << "..." << std::endl;
            } else {
                std::cout << toolName << " is already up to date." << std::endl;
                return 0;
            }
        } else {
            // If we can't get remote hash, assume update is needed
            needsUpdate = true;
            std::cout << "Updating " << toolName << "..." << std::endl;
        }
    } else {
        std::cout << "Installing " << toolName << "..." << std::endl;
    }
    
    // Remove existing installation
    DropshellScriptManager scriptManager;
    scriptManager.removeToolEntry(toolName);
    if (std::filesystem::exists(binDir)) std::filesystem::remove_all(binDir);
    
    // Download tool
    GetbinClient getbin2;
    std::cout << "Downloading " << toolName << ":" << arch << "..." << std::endl;
    if (!getbin2.download(toolName, arch, archivePath.string())) {
        std::cerr << "Failed to download tool archive." << std::endl;
        return 1;
    }
    
    // Unpack tool
    ArchiveManager archiver;
    if (!archiver.unpack(archivePath.string(), binDir.string())) {
        std::cerr << "Failed to unpack tool archive." << std::endl;
        return 1;
    }
    
    // Clean up the archive file
    std::filesystem::remove(archivePath);
    
    // Add to PATH and autocomplete
    scriptManager.addToolEntry(toolName, binDir.string());
    scriptManager.addAutocomplete(toolName);
    
    // Get tool info
    std::string hash;
    getbin2.getHash(toolName, arch, hash);
    std::string version;
    std::string toolPath = binDir.string() + "/" + toolName;
    std::string versionCmd = shellEscape(toolPath) + " version";
    FILE* fp = popen(versionCmd.c_str(), "r");
    if (fp) {
        char buf[128];
        if (fgets(buf, sizeof(buf), fp)) {
            version = std::string(buf);
            // Remove trailing newline if present
            if (!version.empty() && version.back() == '\n') {
                version.pop_back();
            }
        }
        pclose(fp);
    } else {
        std::cerr << "Warning: Failed to get version for " << toolName << std::endl;
    }
    
    // Save tool info
    json toolInfo = {
        {"name", toolName},
        {"version", version},
        {"hash", hash},
        {"arch", arch}
    };
    std::ofstream toolInfoFile(toolInfoPath);
    toolInfoFile << toolInfo.dump(2);
    toolInfoFile.close();
    
    // Run setup script if exists
    std::filesystem::path setupScriptPath = binDir / "setup_script.sh";
    if (std::filesystem::exists(setupScriptPath)) {
        std::cout << "Running setup script..." << std::endl;
        std::string cmd = setupScriptPath.string();
        std::system(cmd.c_str());
    }
    
    std::cout << "Installed " << toolName << " successfully." << std::endl;
    return 0;
}

int publish_tool(int argc, char* argv[]) {
    if (argc < 4) {
        std::cerr << "Usage: getpkg publish <tool_name:ARCH> <folder>" << std::endl;
        return 1;
    }
    std::string labeltag = argv[2];
    std::string folder = argv[3];
    std::string home = get_home();
    std::filesystem::path archivePath = std::filesystem::path(home) / ".tmp" / (labeltag + ".tgz");
    std::filesystem::create_directories(archivePath.parent_path());

    ArchiveManager archiver;
    if (!archiver.pack(folder, archivePath.string())) {
        std::cerr << "Failed to create archive." << std::endl;
        return 1;
    }
    std::string token;
    const char* envToken = std::getenv("SOS_WRITE_TOKEN");
    if (envToken && std::strlen(envToken) > 0) {
        token = envToken;
    } else {
        std::filesystem::path tokenPath = std::filesystem::path(home) / ".config/tools.dropshell.app/write_token.txt";
        if (std::filesystem::exists(tokenPath)) {
            std::ifstream tfile(tokenPath);
            std::getline(tfile, token);
        } else {
            std::cout << "Enter tools.dropshell.app write token: ";
            std::getline(std::cin, token);
            std::filesystem::create_directories(tokenPath.parent_path());
            std::ofstream tfile(tokenPath);
            tfile << token << std::endl;
        }
    }
    GetbinClient getbin;
    std::string url, hash;
    if (!getbin.upload(archivePath.string(), url, hash, token)) {
        std::cerr << "Failed to upload archive." << std::endl;
        return 1;
    }
    std::cout << "Published! URL: " << url << "\nHash: " << hash << std::endl;
    return 0;
}

int update_tool(int argc, char* argv[]) {
    std::cout << "Updating getpkg itself..." << std::endl;
    
    // First update getpkg itself
    char* fakeArgv[] = {argv[0], (char*)"install", (char*)"getpkg"};
    int result = install_tool(3, fakeArgv);
    if (result != 0) {
        std::cerr << "Failed to update getpkg" << std::endl;
        return result;
    }
    
    // Then update all installed tools
    std::string home = get_home();
    std::filesystem::path configDir = std::filesystem::path(home) / ".config/getpkg";
    std::cout << "Updating all installed tools..." << std::endl;
    
    for (const auto& entry : std::filesystem::directory_iterator(configDir)) {
        if (entry.path().extension() == ".json") {
            std::string tname = entry.path().stem();
            if (tname != "getpkg") {  // Skip getpkg since we already updated it
                char* toolArgv[] = {argv[0], (char*)"install", (char*)tname.c_str()};
                install_tool(3, toolArgv);
            }
        }
    }
    
    std::cout << "Update complete." << std::endl;
    return 0;
}

int create_tool(int argc, char* argv[]) {
    if (argc < 4) {
        std::cerr << "Usage: getpkg create <tool_name> <directory_name>" << std::endl;
        return 1;
    }
    std::string toolName = argv[2];
    std::string directoryName = argv[3];
    std::filesystem::path toolDir = std::filesystem::current_path() / directoryName;
    if (!std::filesystem::exists(toolDir)) {
        std::filesystem::create_directories(toolDir);
        std::cout << "Created directory: " << toolDir << std::endl;
    } else {
        std::cout << "Directory already exists: " << toolDir << std::endl;
    }
    std::filesystem::path setupScriptPath = toolDir / "setup_script.sh";
    if (!std::filesystem::exists(setupScriptPath)) {
        std::ofstream setupFile(setupScriptPath);
        setupFile << "#!/bin/bash\necho 'Setup complete.'\n";
        setupFile.close();
        std::filesystem::permissions(setupScriptPath, std::filesystem::perms::owner_exec | std::filesystem::perms::owner_write | std::filesystem::perms::owner_read);
        std::cout << "Created setup script: " << setupScriptPath << std::endl;
    } else {
        std::cout << "Setup script already exists: " << setupScriptPath << std::endl;
    }
    return 0;
}

} // end anonymous namespace

int main(int argc, char* argv[]) {
    if (argc < 2) {
        std::cout << "Usage: getpkg <tool_name> | <command> [args...]" << std::endl;
        return 1;
    }
    std::string command = argv[1];
    if (command == "install") {
        return install_tool(argc, argv);
    } else if (command == "publish") {
        return publish_tool(argc, argv);
    } else if (command == "update") {
        return update_tool(argc, argv);
    } else if (command == "autocomplete") {
        std::vector<std::string> args(argv + 2, argv + argc);
        if (args.empty()) std::cout << R"(install
publish
update
version
create
help
)";
    } else if (command == "version") {
        std::cout << dropshell::VERSION << std::endl;
    } else if (command == "create") {
        return create_tool(argc, argv);
    } else if (command == "help") {
        std::cout << "Usage: getpkg <tool_name> | <command> [args...]" << std::endl;
        std::cout << "Commands:" << std::endl;
        std::cout << "  <tool_name>                       Install/update the specified tool" << std::endl;
        std::cout << "  install <tool_name>               Install/update tool (legacy)" << std::endl;
        std::cout << "  publish <tool_name:ARCH> <folder> Publish a tool" << std::endl;
        std::cout << "  update                            Update getpkg and all tools" << std::endl;
        std::cout << "  version                           Show getpkg version" << std::endl;
        std::cout << "  create <tool_name> <directory>    Create new tool directory" << std::endl;
    } else {
        // Assume it's a tool name and install it
        char* fakeArgv[] = {argv[0], (char*)"install", argv[1]};
        return install_tool(3, fakeArgv);
    }
    return 0;
}